const Base64 = require("../functions/Base64");
const UpdaterProdConfig = require("../model/UpdaterProdConfig");
const ConfigProviderUnitCasher = require("../provider/UpdaterProductConfig");
const AssistantControl = require("./AssistantControl");

class ProductConfig {

  async getConfigByUnitCasher(params, { latest = false } = {}) {

    const provider = ConfigProviderUnitCasher;
    const method = latest ? provider.getLatestProdConfig.bind(provider) : provider.getProductConfig.bind(provider);
    const { rows: [ config ]} = await method(params);

    if (config) {
      config.CONFIG = Base64.decode(config.CONFIG);
      return config;
    }
    return null;
  }

  async getConfig( params, { latest = false } = {}){
    const NRVERSAO = latest ? null : params.NRVERSAO;
    const { rows: [ config ]} = await UpdaterProdConfig.getLatestProdConfig({ ...params, NRVERSAO });

    if (config) {
      config.CONFIG = Base64.decode(config.CONFIG);
      return config;
    }
    return null;
  }

  mergeConfig(originalConfig, newConfig) {
    let changed = false;
    let mergedConfig = {};
    if (!newConfig) {
      return { config: originalConfig, changed };
    }
    newConfig.forEach(({ name, value }) => {
      const hasProp = (name in originalConfig);
      const currentValue = originalConfig[name];
      if (!hasProp || (value && !currentValue)) {
        changed = changed || (!hasProp || value !== currentValue);
        mergedConfig[name] = value;
      } else {
        mergedConfig[name] = originalConfig[name];
      }
    });
    return { config: mergedConfig, changed }
  }

  async handleConfigInDataBase(baseConfig, controlConfig, action, params) {
    const defaultConfig = (baseConfig || {}).CONFIG || {};
    const { changed, config } = this.mergeConfig(defaultConfig, controlConfig);
    if (!changed && action === 'update') return { changed, config };
    const CONFIG = Base64.encode(config);
    if (action === 'create') {
      await ConfigProviderUnitCasher.insert({ ...params, CONFIG });
    } else if (action === 'update') {
      await ConfigProviderUnitCasher.update({ ...params, CONFIG });
    }
    return { changed, config }
  }

  handleConfigData(configData, controlConfig) {
    let config = {};
    let hasEmptyValue = false;
    const defaultLockedParams = ['CONFIG_FILE'];

    const emptyProps = {
      label: '',
      description: '',
      isVisible: true,
      isEditable: true,
    };
    for (const paramName in configData) {
      const controlParam = (controlConfig || []).find(({ name }) => name === paramName) || emptyProps;
      let value = configData[paramName].toString();
      config[paramName] = {
        ...controlParam,
        name: paramName,
        value,
      }
      config[paramName].isEditable = config[paramName].isEditable && !defaultLockedParams.includes(paramName);
      if (!value && config[paramName].isVisible) {
        hasEmptyValue = true;
      }
    }

    return { config, hasEmptyValue };
  }

  async getProdConfig(requestParams) {
    const {
      UPDATERPRODID: CDPRODUCTID,
      NRVERSAOATUAL: NRVERSAO,
      CDFILIAL,
      CDCAIXA = null,
    } = requestParams;

    try {
      const params = { CDPRODUCTID, NRVERSAO, CDFILIAL, CDCAIXA };

      if (!params.CDPRODUCTID) {
        throw new Error('Codigo do produto não foi informado.');
      }

      const controlConfig = await AssistantControl.getProductSettings(CDPRODUCTID);
      const currentConfig = await this.getConfigByUnitCasher(params);
      let config = {};
      if (currentConfig) {
        const result = await this.handleConfigInDataBase(currentConfig, controlConfig, 'update', params);
        config = result.config;
      } else {
        const latestConfig = await this.getConfigByUnitCasher(params, { latest: true });
        const result = await this.handleConfigInDataBase(latestConfig, controlConfig, 'create', params);
        config = result.config;
      }

      const response = {
        error: false,
        message: '',
        errorInfo: {},
      }

      const { config: configData, hasEmptyValue } = this.handleConfigData(config, controlConfig);
      response.data = configData;
      if (hasEmptyValue) {
        response.message = 'Preencha os parâmetros do projeto antes de continuar instalação.';
      }

      return response;
    } catch (error) {
      return {
        error: true,
        message: error.message
      }
    }
  }
}

module.exports = new ProductConfig();