const { v4: uuidv4 } = require("uuid");

const updaterModel = require("../model/updaterModel");
const updaterModelCasher = require('../model/updaterModelCasher');
const UpdaterEvents = require('../provider/UpdaterEvents');
const GeraNovoCodigo = require("../modules/NovoCodigoModule");
const Base64 = require("../functions/Base64");
const ServicesMonitoring = require('../helpers/ServicesMonitoring');
const SystemLogger = require("../logs/SystemLogger");
const UpdaterServiceMonitoring = require("../provider/UpdaterServiceMonitoring");
const UpdaterHistory = require("../provider/UpdaterHistory");
const UpdaterUnitProduct = require("../provider/UpdaterUnitProduct");
const DB = require('../database/DB');
const Crypt = require('./Crypt');
const UpdaterConfig = require('../provider/UpdaterConfig');
const UpdaterConfigHist = require("../provider/UpdaterConfigHist");
const UpdaterProdConfig = require("../model/UpdaterProdConfig");
const UpdaterProductConfig = require("../provider/UpdaterProductConfig");
const ProductConfig = require("./ProductConfig");

class UpdaterService {
  updaterLog = {};

  async getHistorySeq(CDCAIXA) {
    return CDCAIXA ? GeraNovoCodigo.prxUpdaterHistoricoCx() : GeraNovoCodigo.prxUpdaterHistorico();
  }

  async getEventSeq(CDCAIXA) {
    return CDCAIXA ? GeraNovoCodigo.prxUpdaterEventoCx() : GeraNovoCodigo.prxUpdaterEvento();
  }

  getRequestIp(req) {
    if (process.env.NODE_ENV === 'development') {
      return 'Ambiente de Desenvolvimento';
    }

    return req.headers['x-forwarded-for'] || req.socket.remoteAddress || `IP não encontrado na requisição`;
  }

  async updateAnalyzeStatus(prodFiliId, status = 'PENDENTE') {
    return updaterModel.updateProdFiliStatus({
      PRODFILIID: prodFiliId,
      STATUS: status,
    });
  }

  async updateResetIntegrityAnalyze(prodFiliId) {
    try {
      await updaterModel.updateResetIntegrityAnalyze({
        PRODFILIID: prodFiliId,
      });
    } catch(err) {
      return Promise.reject(err);
    }
  }

  async getProductsToIntegrityAnalize(unit) {
    return updaterModel.getProductsToIntegrityAnalize({
      CDFILIAL: unit,
    });
  }

  async insertEventTable(data, conn = null) {
    const {
      CDOPERADOR,
      IDSTEVENTO,
      CDFILIAL,
      CDCAIXA = null,
      DEVOPSID = null,
      PRODFILIID = null,
      CURRENTVERSION = null,
      INSTALLVERSION = null,
      CREATENEWVERSION = 'N',
    } = data;

    let response = {
      error: true,
      message: "SYSTEM_LOGGED_EVENT",
    };

    const autoTransactionControl = !conn;

    let params = {
      P_CDFILIAL: CDFILIAL,
      P_CDCAIXA: CDCAIXA,
      P_CDPRODUCTID: DEVOPSID,
    };

    const unitEvents = await UpdaterEvents.getActiveEvents(params);

    if (unitEvents.rows.length > 0) return response;

    const eventSeq = await this.getEventSeq(CDCAIXA);
    const uuid = uuidv4();
    const basicParams = {
      P_CDFILIAL: CDFILIAL,
      P_CDCAIXA: CDCAIXA,
      P_CDEVENTO: eventSeq,
      P_CDUUID: uuid,
      P_CDPRODUCTID: DEVOPSID,
      P_IDSTEVENTO: IDSTEVENTO,
      P_NRVERSAOATUAL: CURRENTVERSION,
      P_NRVERSAOALVO: INSTALLVERSION,
      P_IDCRIARVERSAO: CREATENEWVERSION,
    };

    let connection = conn;
    if(!connection) {
      connection = await DB.beginTransaction();
    }
    try {
      await UpdaterEvents.insert({...basicParams, P_CDPRODFILIID: PRODFILIID}, { connection });

      const historySeq = await this.getHistorySeq(CDCAIXA);

      const historyParams = {
        P_CDHISTORICO: historySeq,
        P_CDOPERADOR: CDOPERADOR,
        P_NMPRODUCT: null,
      };

      await UpdaterHistory.insert({...basicParams, ...historyParams}, { connection });

      response = {
        error: false,
        message: "EVENT_REGISTERED_SUCCESSFULLY",
      };
      if(autoTransactionControl) {
        connection.commit();
      }
    } catch(err) {
      if(autoTransactionControl) {
        connection.rollback();
      }
    } finally {
      if(autoTransactionControl) {
        connection.close();
      }
    }
    return response;
  }

  async updateEventTable(data, conn = null) {
    const {
      CDFILIAL,
      CDCAIXA = null,
      CDOPERADOR,
      DEVOPSID = null,
      CURRENTVERSION = null,
      INSTALLVERSION = null,
      CREATENEWVERSION = 'N',
      OPERATION
    } = data;

    let response = {
      error: true,
      message: "O evento já foi atualizado no sistema.",
    };
    const autoTransactionControl = !conn;
    let eventId;

    switch (OPERATION) {
      case 'INSTALACAO_EM_CURSO':
        eventId = 'INSTALACAO_PENDENTE';
      break;
      case 'REMOCAO_EM_CURSO':
        eventId = 'REMOCAO_PENDENTE';
      break;
      case 'VERIFICACAO_EM_CURSO':
        eventId = 'VERIFICACAO_PENDENTE';
      break;
      case 'ANALISE_INTEGRIDADE_EM_CURSO':
        eventId = 'ANALISE_INTEGRIDADE_PENDENTE'
    }

    let params = {
      P_CDFILIAL: CDFILIAL,
      P_CDCAIXA: CDCAIXA,
      P_CDPRODUCTID: DEVOPSID,
      P_IDSTEVENTO: eventId,
    };

    const pendingEvents = await UpdaterEvents.getProductEvents(params);

    if (pendingEvents.rows.length > 0) {
      const uuid = uuidv4();

      params = {
        P_CDCAIXA: CDCAIXA,
        P_IDSTEVENTO: OPERATION,
        P_CDEVENTO: pendingEvents.rows[0].CDEVENTO,
      };

      let connection = conn;
      if(!connection) {
        connection = await DB.beginTransaction();
      }
      try {
        await UpdaterEvents.updateEventStatus(params, { connection });

        const history = await UpdaterHistory.getEventHistory(params);

        if (history.rows.length === 0) {

          const historySeq = await this.getHistorySeq(CDCAIXA);

          params = {
            P_CDFILIAL: CDFILIAL,
            P_CDCAIXA: CDCAIXA,
            P_CDHISTORICO: historySeq,
            P_CDEVENTO: pendingEvents.rows[0].CDEVENTO,
            P_CDUUID: uuid,
            P_CDOPERADOR: CDOPERADOR,
            P_CDPRODUCTID: DEVOPSID,
            P_NMPRODUCT: null,
            P_IDSTEVENTO: OPERATION,
            P_NRVERSAOATUAL: CURRENTVERSION,
            P_NRVERSAOALVO: INSTALLVERSION,
            P_IDCRIARVERSAO: CREATENEWVERSION,
          };

          await UpdaterHistory.insert(params, { connection });

          response = {
            error: false,
            message: "Evento atualizado com sucesso. Histórico criado",
          };
          if(autoTransactionControl) {
            connection.commit();
          }
        }
      } catch(err) {
        if(autoTransactionControl) {
          connection.rollback();
        }
      } finally {
        if(autoTransactionControl) {
          connection.close();
        }
      }
    }

    return response;
  }

  async finishEventTable(data, conn = null) {
    const {
      CDFILIAL,
      CDCAIXA = null,
      CDOPERADOR,
      DEVOPSID = null,
      PRODFILIID = null,
      CURRENTVERSION = null,
      INSTALLVERSION = null,
      CREATENEWVERSION = 'N'
    } = data;

    let response = {
      error: true,
      message: "O evento já foi atualizado no sistema.",
    };

    const autoTransactionControl = !conn;

    let params = {
      P_CDFILIAL: CDFILIAL,
      P_CDCAIXA: CDCAIXA ? `_${CDCAIXA}_` : null,
    };

    const inProgressEvents = await UpdaterEvents.getEventsInProgress(params);

    if (inProgressEvents.rows.length > 0) {
      params = {
        P_CDCAIXA: CDCAIXA ? `_${CDCAIXA}_` : null,
        P_IDSTEVENTO: "FINALIZADO",
        P_CDEVENTO: inProgressEvents.rows[0].CDEVENTO,
      };

      let connection = conn;
      if(!connection) {
        connection = await DB.beginTransaction();
      }
      try {
        await UpdaterEvents.updateEventStatus(params, { connection });

        const history = await UpdaterHistory.getEventHistory(params);

        if (history.rows.length === 0) {
          const uuid = uuidv4();
          const historySeq = await this.getHistorySeq(CDCAIXA);

          params = {
            P_CDFILIAL: CDFILIAL,
            P_CDCAIXA: CDCAIXA,
            P_CDHISTORICO: historySeq,
            P_CDEVENTO: inProgressEvents.rows[0].CDEVENTO,
            P_CDUUID: uuid,
            P_CDOPERADOR: CDOPERADOR,
            P_CDPRODUCTID: DEVOPSID,
            P_NMPRODUCT: null,
            P_IDSTEVENTO: "FINALIZADO",
            P_NRVERSAOATUAL: CURRENTVERSION,
            P_NRVERSAOALVO: INSTALLVERSION,
            P_IDCRIARVERSAO: CREATENEWVERSION,
          };

          await UpdaterHistory.insert(params, { connection });

          if (PRODFILIID != null) {
            await UpdaterUnitProduct.update({
              P_CDCAIXA: CDCAIXA,
              P_ID: PRODFILIID,
              P_NRVERSAO: INSTALLVERSION
            }, { connection })
          }

          response = {
            error: false,
            message: "Evento atualizado com sucesso. Histórico criado",
          };
          if(autoTransactionControl) {
            connection.commit();
          }
        }
      } catch(err) {
        if(autoTransactionControl) {
          connection.rollback();
        }
      } finally {
        if(autoTransactionControl) {
          connection.close();
        }
      }
    }

    return response;
  }

  async cancelUnitEvents(data) {
    const {
      CDFILIAL,
      CDCAIXA = null
    } = data;

    const params = {
      P_CDFILIAL: CDFILIAL,
      P_CDCAIXA: CDCAIXA
    }

    await UpdaterEvents.cancelEvent(params);

  }

  async cancelEventTable(data) {
    const {
      CDFILIAL,
      CDCAIXA,
      CDOPERADOR,
      DEVOPSID,
      CURRENTVERSION,
      INSTALLVERSION,
      CREATENEWVERSION
    } = data;

    let response = {
      error: true,
      message: "O evento já foi atualizado no sistema.",
    };

    let params = {
      P_CDFILIAL: CDFILIAL,
      P_CDCAIXA: CDCAIXA,
      P_CDPRODUCTID: DEVOPSID,
    };

    const unitEvents = await UpdaterEvents.getActiveEvents(params);

    if (unitEvents.rows.length > 0) {
      const uuid = uuidv4();

      params = {
        P_CDCAIXA: CDCAIXA,
        P_IDSTEVENTO: "CANCELADO",
        P_CDEVENTO: unitEvents.rows[0].CDEVENTO,
      };

      const connection = await DB.beginTransaction();

      try {

        await UpdaterEvents.updateEventStatus(params, { connection });

        const history = await UpdaterHistory.getEventHistory(params);

        if (history.rows.length === 0) {
          const historySeq = await this.getHistorySeq(CDCAIXA);

          params = {
            P_CDFILIAL: CDFILIAL,
            P_CDCAIXA: CDCAIXA,
            P_CDHISTORICO: historySeq,
            P_CDEVENTO: unitEvents.rows[0].CDEVENTO,
            P_CDUUID: uuid,
            P_CDOPERADOR: CDOPERADOR,
            P_CDPRODUCTID: DEVOPSID,
            P_NMPRODUCT: null,
            P_IDSTEVENTO: "CANCELADO",
            P_NRVERSAOATUAL: CURRENTVERSION,
            P_NRVERSAOALVO: INSTALLVERSION,
            P_IDCRIARVERSAO: CREATENEWVERSION,
          };

          await UpdaterHistory.insert(params, { connection });

          response = {
            error: false,
            message: "Evento atualizado com sucesso. Histórico criado",
          };

          connection.commit();
        }
      } catch(err) {
        connection.rollback();
      } finally {
        connection.close();
      }
    }

    return response;
  }

  async getPendingEventsByUnit(CDFILIAL, CDCAIXA) {
    try {
      const params = {
        P_CDFILIAL: CDFILIAL,
        P_CDCAIXA: CDCAIXA,
      };
      const response = {
        error: false,
        message: 'Eventos pendentes encontrados:',
        events: [],
        products: [],
      };

      const model = CDCAIXA ? updaterModelCasher : updaterModel;
      const pendingEvents = await UpdaterEvents.getPendingEvents(params);
      const hasIntegrityAnalyzeEvent = pendingEvents.rows.some(event => event.IDSTEVENTO === 'ANALISE_INTEGRIDADE_PENDENTE');

      if(hasIntegrityAnalyzeEvent && model.getProductsToIntegrityAnalize) {
        const { rows } = await model.getProductsToIntegrityAnalize({ CDFILIAL });
        response.products = rows;
      }

      response.events = pendingEvents.rows;
      return response;
    } catch (error) {
      SystemLogger.write(error);
      return {
        error: true,
        message: error.message,
        events: []
      }
    }
  }

  async checkVersion(row) {
    const {
      CDFILIAL: P_CDFILIAL,
      CDCAIXA: P_CDCAIXA,
      UPDATERPRODID: P_UPDATERPRODID,
      NRVERSAOATUAL: P_NRVERSAOATUAL
    } = row;
    const params = { P_CDFILIAL, P_CDCAIXA, P_UPDATERPRODID, P_NRVERSAOATUAL };
    const result = await UpdaterUnitProduct.getVersion(params);
    return { error: result.rows.length > 0 };
  }

  async getConfigIniOnDatabase(config) {
    try {
      const { CDFILIAL, CDCAIXA } = config;
  
      const params = {
        P_CDFILIAL: CDFILIAL,
        P_CDCAIXA: CDCAIXA,
      };
  
      const result = await UpdaterConfig.getConfigIni(params);
  
      if (result.rows.length === 0) {
        return '';
      };
  
      const configDb = result.rows.shift();
      return Base64.decode(configDb.CONFIG);
    } catch (error) {
      SystemLogger.error(error);
      throw error;
    }
  }

  async loadConfig({ config }) {
    try {
      const configIni = await this.getConfigIniOnDatabase(config.environment);

      if (configIni === '') {
        const { CDFILIAL, CDCAIXA } = config.environment;
        const message = CDCAIXA ? `filial ${CDFILIAL} e caixa ${CDCAIXA}` : `filial ${CDFILIAL}`;

        return {
          error: true,
          message: `Arquivo config.ini da ${message} não foi encontrado na base de dados.`
        };
      }

      return {
        error: false,
        data: configIni
      }
    } catch(error) {
      SystemLogger.error(error);

      return error;
    }
  }

  mergeObjects(target, source) {
    for (const key in source) {
      if (source.hasOwnProperty(key)) {
        if (source[key] instanceof Object && key in target) {
          this.mergeObjects(target[key], source[key]);
        } else {
          target[key] = source[key];
        }
      }
    }
  }

  async updateConfig({ config, info }){
    const connection = await DB.beginTransaction();
    let response = {};
    try {
      const { CDFILIAL, CDCAIXA } = info;

      if (!CDFILIAL) {
        throw {
          error: true,
          message: 'Codigo da filial não foi informado'
        };
      }

      const params = {
        P_CDFILIAL: CDFILIAL,
        P_CDCAIXA: CDCAIXA,
      };

      const message = CDCAIXA ? `filial ${CDFILIAL} e caixa ${CDCAIXA}` : `filial ${CDFILIAL}`;

      const configIni = await this.getConfigIniOnDatabase(info);

      if (configIni === '') {
        throw {
          error: true,
          message: `Arquivo config.ini da ${message} não foi encontrado na base de dados.`
        };
      }

      this.mergeObjects(configIni, config);

      if ('assistant' in configIni && 'FORCED_SYNC' in configIni.assistant) {
        delete configIni.assistant.FORCED_SYNC;
      }

      params.P_CONFIG = Base64.encode(configIni);

      const updateConfig = await UpdaterConfig.update(params, { connection });

      if (updateConfig.rowsAffected === 0 || updateConfig.outBinds.ID.length === 0) {
        throw {
          error: true,
          message: `Erro ao atualizar as informações do config.ini da ${message} na base de dados.`
        };
      }

      params.P_CDOPERADOR = await this.getDefaultOperator(params);
      params.P_UPDATERCONFIGID = updateConfig.outBinds.ID[0];

      const updateConfigHist = await UpdaterConfigHist.insert(params, { connection });

      if (updateConfigHist.rowsAffected === 0) {
        throw {
          error: true,
          message: `Falha ao atualizar na tabela de histórico de alteração do arquivo config.ini da ${message}.`
        };
      }
      
      connection.commit();

      response = {
        error: false,
        message: `Config.ini da ${message} foi atualizado com sucesso.`
      }
    } catch(error) {
      connection.rollback();
      SystemLogger.error(error);

      response = error;
    } finally {
      connection.close();
    }
    return response;
  }

  isSyncForced(config) {
    if (!('assistant' in config)) {
      return false;
    }

    if (!('FORCED_SYNC' in config.assistant)) {
      return false;
    }

    const { FORCED_SYNC } = config.assistant;

    if (!!FORCED_SYNC) {
      return true;
    }

    return false;
  }

  async saveConfig({ config }){
    const connection = await DB.beginTransaction();
    let response = {};
    try {
      if (!config) {
        throw {
          error: true,
          message: 'Codigo da Filial não foi informado'
        };
      }

      const { CDFILIAL, CDCAIXA } = config.environment;

      const message = CDCAIXA ? `filial ${CDFILIAL} e caixa ${CDCAIXA}` : `filial ${CDFILIAL}`;
      
      const configIni = await this.getConfigIniOnDatabase(config.environment);
      
      if (configIni !== '') {
        if (this.isSyncForced(config)) {
          const update = { config, info: config.environment };
          await this.updateConfig(update);
          return {
            error: false,
            message: `Arquivo config.ini da ${message} foi atualizado na base de dados.`
          }
        } else {
          return {
            error: false,
            message: `Arquivo config.ini da ${message} já está salvo na base de dados.`
          };
        }
      }

      delete config.assistant.FORCED_SYNC;
      const configIniBase64 = Base64.encode(config);

      if (!configIniBase64) {
        throw {
          error: true,
          message: `Não foi possivel converter config.ini da ${message} para base64.`
        };
      }

      const params = {
        P_CDFILIAL: CDFILIAL,
        P_CDCAIXA: CDCAIXA,
        P_CONFIG: configIniBase64
      };

      const insertConfig = await UpdaterConfig.insert(params, { connection });

      if (insertConfig.rowsAffected === 0 || insertConfig.outBinds.ID.length === 0) {
        throw {
          error: true,
          message: `Erro ao inserir as informações do config.ini da ${message} na base de dados.`
        };
      }

      params.P_CDOPERADOR = await this.getDefaultOperator(params);
      params.P_UPDATERCONFIGID = insertConfig.outBinds.ID[0];

      const updateConfigHist = await UpdaterConfigHist.insert(params, { connection });

      if (updateConfigHist.rowsAffected === 0) {
        throw {
          error: true,
          message: `Falha ao inserir na tabela de histórico de alteração do arquivo config.ini da ${message}.`
        };
      }
      connection.commit();

      response = {
        error: false,
        message: `Config.ini da ${message} foi inserido com sucesso.`
      }
    } catch(error) {
      connection.rollback();
      SystemLogger.error(error);

      response = error;
    } finally {
      connection.close();
    }
    return response;
  }

  async getDefaultOperator(params) {
    try {
      const operator = await updaterModel.getDefaultOperator(params);
  
      if (operator.rows.length === 0) {
        throw {
          error: true,
          message: `Não foi possivel buscar o operador/vendedor padrão da filial ${params.P_CDFILIAL}.`
        };
      }
  
      if (operator.rows[0].CDOPERADOR === null && operator.rows[0].CDVENDEDOR === null) {
        throw {
          error: true,
          message: `Os campos de Operador Padrão e Vendedor Padrão da filial ${params.P_CDFILIAL} não estão preenchidos.`
        };
      }
  
      const { CDOPERADOR } = operator.rows[0];
      const defaultOperator = CDOPERADOR ? CDOPERADOR : await updaterModel.getDefaultOperatorFromSalesman({P_CDVENDEDOR: operator.rows[0].CDVENDEDOR}).then(response => {
        if(response && response.rows && response.rows[0]) {
          let [ firstRow ] = response.rows;
          let { CDOPERADOR } = firstRow;
          return CDOPERADOR;
        }
        return null;
      });
  
      return defaultOperator;
    } catch (error) {
      throw error;
    }
  }

  async getDefaultOperatorFromSalesman(salesmanCode) {
    try {
      params.P_CDVENDEDOR = salesmanCode;
      const salesman = await updaterModel.getDefaultOperatorFromSalesman(params);

      if (salesman.rows.length === 0 || salesman.rows[0].CDOPERADOR !== null) {
        throw {
          error: true,
          message: `O campo de operador padrão deve estar preenchido ou associado a um vendedor padrão para a filial ${params.P_CDFILIAL}.`
        };
      }

      return salesman.rows[0].CDOPERADOR
    } catch (error) {
      throw error;
    }
  }

  async getConfigIniData(params) {
    try {
      const result = await UpdaterConfig.getConfigIni(params);

      if (result.rows.length === 0) {
        throw new Error(`Não existe um arquivo de configuração salvo para a filial ${params.P_CDFILIAL}`);
      }

      return result.rows[0];
    } catch (error) {
      throw error;
    }
  }

  async getConfigIniByUnit(cdFilial) {
    try {
      const params = {
        P_CDFILIAL: cdFilial,
      };

      const result = await this.getConfigIniData(params);

      const decodeConfig = Base64.decode(result.CONFIG);

      return {
        error: false,
        message: 'SUCCESS_LOAD_CONFIG_INI',
        data: decodeConfig
      }
    } catch (error) {
      return {
        error: true,
        message: 'FAIL_LOAD_CONFIG_INI'
      }
    }
  }

  async decodeConfigPassword(config) {
    const { PASSWORD, DBENCRYPTED } = config.database;

    if (!DBENCRYPTED) {
      return PASSWORD;
    }

    const { DB_CRYPT_KEY } = process.env;
      
    return Crypt.decrypt(PASSWORD, DB_CRYPT_KEY);
  }

  async encodeConfigPassword(config) {
    const { PASSWORD } = config.database;
    const { DB_CRYPT_KEY } = process.env;
      
    return Crypt.encrypt(PASSWORD, DB_CRYPT_KEY);
  }

  async getConfigIni({ CDFILIAL, CDCAIXA }) {
    try {
      const params = {
        P_CDFILIAL: CDFILIAL,
        P_CDCAIXA: CDCAIXA
      }
  
      const result = await UpdaterConfig.getConfigIni(params);

      if(!result.rows.length) {
        throw 'Nenhum arquivo de configuração encontrado';
      }

      let config = Base64.decode(result.rows[0].CONFIG);

      config.database.PASSWORD = await this.decodeConfigPassword(config);

      return {
        error: false,
        message: 'SUCCESS_LOAD_CONFIG_INI',
        data: config
      }
    } catch (error) {
      return {
        error: true,
        message: 'FAIL_LOAD_CONFIG_INI'
      }
    }
  }

  async saveConfigIni({ CDOPERADOR, CDFILIAL, CDCAIXA, CONFIG }) {
    const connection = await DB.beginTransaction();
    let response = {};

    try {
      let result;

      const encodedPassword = await this.encodeConfigPassword(CONFIG);
      CONFIG.database.PASSWORD = encodedPassword;

      let params = {
        P_CDOPERADOR: CDOPERADOR,
        P_CDFILIAL: CDFILIAL,
        P_CDCAIXA: CDCAIXA,
        P_CONFIG: Base64.encode(CONFIG)
      }

      result = await UpdaterConfig.update(params, { connection });

      if (result.rowsAffected === 0) {
        throw new Error('NO_CONFIG_FILE_FOR_SPECIFIC_UNIT');
      }

      result = await this.getConfigIniData(params);

      params.P_UPDATERCONFIGID = result.ID;

      result = await UpdaterConfigHist.insert(params, { connection });

      if (result.rowsAffected === 0) {
        throw new Error('FAIL_SAVE_CONFIG_INI_HISTORY');
      }

      connection.commit();
      response = {
        error: false,
        message: 'SUCCESS_SYNC_CONFIG_INI',
        params: { 'PASSWORD': encodedPassword }
      }
    } catch (error) {
      connection.rollback();

      response = {
        error: true,
        message: error.message,
        params: { 'CDFILIAL': cdFilial }
      }
    } finally {
      connection.close();
      return response;
    }
  }

  async getConfigHistory({CDFILIAL, CDCAIXA}) {
    try {
      let result;

      const params = {
        P_CDFILIAL: CDFILIAL,
        P_CDCAIXA: CDCAIXA,
      };

      result = await this.getConfigIniData(params);

      params.P_ID = result.ID;

      result = await UpdaterConfigHist.getConfigHist(params);

      return {
        error: false,
        message: 'SUCCESS_LOAD_UPDATE_HISTORY_CONFIG',
        data: result.rows
      }
    } catch (error) {
      return {
        error: true,
        message: error.message
      }
    }
  }

  async saveProdConfig(params){
    const { CDPRODUCTID, NRVERSAO, CONFIG, CDFILIAL, CDCAIXA = null } = params;
    try {
      const hasEmptyValue = Object.values(CONFIG).some( value => value === '');
      if(hasEmptyValue) throw new Error('Ocorreu um erro ao salvar as informações, tente novamente.');

      const encodedConfig = typeof CONFIG === 'string' ? CONFIG : Base64.encode(CONFIG);

      if (!encodedConfig) {
        throw new Error(`Não foi possivel converter os parâmetros do produto ${CDPRODUCTID} para base64.`);
      }

      const updaterProductConfig = {
        CDFILIAL,
        CDCAIXA,
        CDPRODUCTID,
        NRVERSAO,
        CONFIG: encodedConfig
      }
      const { rows } = await UpdaterProductConfig.getProductConfig({ CDPRODUCTID, NRVERSAO, CDFILIAL, CDCAIXA });
      let response = null;
      if(rows && rows.length) {
        response = await UpdaterProductConfig.update(updaterProductConfig);
      } else {
        response = await UpdaterProductConfig.insert(updaterProductConfig);
      }

      if (response && response.rowsAffected === 0) {
        throw new Error(`Falha ao salvar parâmetros do produto ${CDPRODUCTID}.`);
      }

      return {
        error: false,
        message: 'Parâmetros do projeto salvos com sucesso!',
        data: CONFIG
      };
    } catch(error) {

      return {
        error: true,
        message: error.message
      }
    }
  }

  handleAddServicesInUnit(actualServices, servicesToAdd, separator = ';') {
    const hasNewServices = servicesToAdd.some( service => !actualServices.includes(service) );
    if(!hasNewServices) return actualServices.join(separator);
    const allServices = actualServices.concat(servicesToAdd);
    const servicesToSave = [ ...new Set(allServices) ];
    return servicesToSave.join(separator);
  }

  handleRemoveServicesInUnit(actualServices, servicesToRemove, separator = ';') {
    const servicesUpdated = actualServices.filter( service => !servicesToRemove.includes(service) );
    return servicesUpdated.join(separator);
  }

  async handleChangeService(cdFilial, casher, updatedServices, operation) {
    const SUCCESS_RESPONSE = {
      error: false,
      message: 'Monitoramento de serviços atualizado com sucesso!'
    };

    const functionsToHandle = {
      INSTALL: this.handleAddServicesInUnit,
      REMOVE: this.handleRemoveServicesInUnit,
      default: (actualServices) => actualServices,
      ERROR: (error) => {
        SystemLogger.write(error);
        throw new Error('Ocorreram erros ao atualizar os servicos monitorados na unidade');
      }
    }

    try {
      const params = {
        P_CDFILIAL: cdFilial,
        P_CDCAIXA: casher || null,
        P_MONITORAMENTO: updatedServices || null,
        P_NRORG: process.env.NRORG,
        P_NRORGINCLUSAO: process.env.NRORG,
        P_NRORGULTATU: process.env.NRORG,
      };

      const services = await UpdaterServiceMonitoring.getServiceMonitoring(params).catch(functionsToHandle.ERROR);

      if (!services.rows.length) {
        await UpdaterServiceMonitoring.insert(params);
      }

      const finalResult = await UpdaterServiceMonitoring.getServiceMonitoring(params).catch(functionsToHandle.ERROR);
      const actualServices = finalResult.rows.length > 0 ? finalResult.rows[0].MONITORAMENTO || '' : '';

      const functionToCall = functionsToHandle[operation.toUpperCase()] || functionsToHandle.default;
      const servicesToSave = functionToCall.apply(this, [actualServices.split(';'), updatedServices.split(';')]);

      params.P_MONITORAMENTO = ServicesMonitoring.removeMonitoringParamInconsistencies(servicesToSave);

      await UpdaterServiceMonitoring.update(params).catch(functionsToHandle.ERROR);

      return SUCCESS_RESPONSE;
    } catch (error) {
      SystemLogger.write(error);
      return {
        error: true,
        message: error
      };
    }
  }

  async installNewVersion(params) {
    const connection = await DB.beginTransaction();
    let response = {};
    try {
      if(!params.UPDATERPRODID) throw 'Código do produto não informado.';

      const { rows } = await updaterModel.getUpdateLastRecordProduct(params);

      if(rows.length === 0) throw 'Erro ao efetuar busca do produto.';

      if(rows[0].NRVERSAO === params.NRVERSAO) throw 'Essa versão já está instalada.';

      const newVersionObject = {
        UPDATERPRODID: params.UPDATERPRODID,
        NRVERSAO: params.NRVERSAO,
        DSURL: params.DSURL,
        DSDIRETORIO: rows[0].DSDIRETORIO,
        IDATIVO: rows[0].IDATIVO,
        IDDOWNLOADVER: 'N',
        IDAPROVVER: rows[0].IDAPROVVER
      };

      const { rowsAffected } = await updaterModel.setUpdateProdVersion(newVersionObject, { connection });

      if(rowsAffected === 0) throw 'Erro ao inserir informações do produto.';

      const lastConfig = await ProductConfig.getConfig({
        CDPRODUCTID: params.UPDATERPRODID,
        NRVERSAO: params.NRVERSAO
      });

      const prodConfig = {
        CDPRODUCTID: params.UPDATERPRODID,
        NRVERSAO: params.NRVERSAO,
        CONFIG: lastConfig ? lastConfig.CONFIG : null
      };

      const resultSet = await UpdaterProdConfig.insert(prodConfig, { connection });

      if(resultSet.rowsAffected === 0) throw 'Erro ao inserir informações do produto.';

      connection.commit();

      response = {
        error: false,
        message: 'As informações foram salvas com sucesso.'
      }
    } catch (error) {
      connection.rollback();
      response = {
        error: true,
        message: error
      }
    } finally {
      connection.close();
    }
    return response;
  }

  async removeEventsByUnitAndProduct( params ) {
    const connection = await DB.beginTransaction();
    let response = {};
    try {
      await UpdaterHistory.delete(params, { connection });
      await UpdaterEvents.deleteProductEvents(params, { connection });
      connection.commit();
      response = {
        error: false,
        message: ''
      }
    } catch( error ) {
      connection.rollback();
      response =  {
        error: true,
        message: error.message
      }
    } finally {
      connection.close();
    }
    return response;
  }

  async updatePackageURL(params) {
    try {
      const { UPDATERPRODID, NRVERSAO, DSURL, DSDIRETORIO } = params;

      if (!UPDATERPRODID || !NRVERSAO || !DSURL || !DSDIRETORIO) {
        throw 'Missing needed parameters: UPDATERPRODID, NRVERSAO, DSURL, DSDIRETORIO';
      }

      const result = await updaterModel.updatePackageURL(params);

      return {
        error: false,
        message: `${result.rowsAffected} registro atualizado.`
      }
    } catch (error) {
      return {
        error: true,
        message: error
      }
    }
  }

  async getProductsVersion(params) {
    const {
      CDFILIAL: P_CDFILIAL,
      CDCAIXA: P_CDCAIXA,
    } = params;
    const model = P_CDCAIXA ? updaterModelCasher : updaterModel;
    return model.getProductsVersion({P_CDFILIAL, P_CDCAIXA});
  }

  async handleInstalledProducts(params) {
    const {
      CDOPERADOR,
      CDFILIAL,
      PROJECTS,
    } = params;
    const connection = await DB.beginTransaction();
    try {
      await this.updateEventTable({
        CDOPERADOR,
        CDFILIAL,
        OPERATION: 'VERIFICACAO_EM_CURSO',
      }, connection);
      const { rows } = await UpdaterUnitProduct.getInstalledProducts({ CDFILIAL });
      const { rows: allVersions } = await updaterModel.getDownloadedVersion();
      const versionByProduct = allVersions.reduce((versionMap, row) => {
        versionMap[row.UPDATERPRODID] = versionMap[row.UPDATERPRODID] || [];
        versionMap[row.UPDATERPRODID].push(row.NRVERSAO);
        return versionMap;
      }, {});
      const versionsToUpdate = {};
      const productVersion = rows.reduce((productMapping, row) => {
        const productId = row.UPDATERPRODID;
        const prodiFiliId = row.ID;
        productMapping[productId] = productMapping[productId] || {};
        productMapping[productId][prodiFiliId] = row.NRVERSAOATUAL;
        return productMapping;
      }, {});

      const invalidProducts = [];
      PROJECTS.forEach(({id: productId, name, versions}) => {
        if(!Array.isArray(versions)) return;
        const [ installedVersion ] = versions;
        let maxSlots = Number((process.env.MAX_MULTI_VERSION_SLOTS || 3));
        if(isNaN(maxSlots)) {
          maxSlots = 3;
        }
        if(!productVersion[productId]) {
          invalidProducts.push({
            id: productId,
            name: productId,
            reason: 'INVALID_PRODUCT_ID',
            installedVersions: versions,
            savedVersions: []
          });
          return;
        }
        const invalidVersions = versions.filter(version => !versionByProduct[productId].includes(version));
        if(invalidVersions.length) {
          invalidProducts.push({
            id: productId,
            name,
            reason: 'INVALID_INSTALLED_VERSION',
            installedVersions: versions,
            savedVersions: Object.values(productVersion[productId]),
            invalidVersions,
          });
          return;
        }
        if(versions.length > Object.keys(productVersion[productId]).length || versions.length > maxSlots) {
          invalidProducts.push({
            id: productId,
            name,
            reason: versions.length > maxSlots ? 'MAX_SLOT_VERSION_REACHABLE' : 'UNAVAILABLE_SLOTS',
            installedVersions: versions,
            savedVersions: Object.values(productVersion[productId])
          });
          return;
        }
        if(versions.length === 0) {
          Object.keys(productVersion[productId]).forEach(prodfiliid => {
            versionsToUpdate[prodfiliid] = null;
          });
          return;
        }
        if(versions.length === 1) { // product with one version
          const [ savedVersion ] = Object.values(productVersion[productId]);
          if(installedVersion !== savedVersion) {
            const [ PRODFILIID ] = Object.keys(productVersion[productId]);
            versionsToUpdate[PRODFILIID] = installedVersion;
          }
        } else { // multi version
          const versionsCopy = [...versions];
          Object.keys(productVersion[productId]).forEach(prodfiliid => {
            versionsToUpdate[prodfiliid] = versionsCopy.shift() || null;
          });
        }
      });
      const allUpdateVersion = Object.keys(versionsToUpdate).map(prodfiliid => {
        const newVersion = versionsToUpdate[prodfiliid];
        return UpdaterUnitProduct.update({
          P_ID: prodfiliid,
          P_NRVERSAO: newVersion,
        }, { connection })
      });
      await Promise.all(allUpdateVersion);
      await this.finishEventTable({
        CDOPERADOR,
        CDFILIAL,
      }, connection);
      connection.commit();
      return {
        invalidProducts,
      }
    } catch(error) {
      connection.rollback();
      await this.finishEventTable({
        CDOPERADOR,
        CDFILIAL,
      });
      SystemLogger.write(error);
      return Promise.reject(error);
    } finally {
      connection.close();
    }
  }

  async updateAnalyzeIntegrity(payload) {
    try {
      const updatePromise = payload.ANALYZE_RESULT.map(product => {
        const { id, integrity, error } = product;
        const status = integrity ? 'ANALISE_VALIDA' : 'ANALISE_INVALIDA';
        if(id) {
          return updaterModel.updateProdFiliStatus({
            PRODFILIID: id,
            STATUS: error ? 'FALHA_NA_ANALISE' : status,
          })
        } else {
          return Promise.reject({ invalidProduct: product });
        }
      });
      await Promise.all(updatePromise);
    } catch (err) {
      return Promise.reject(err);
    }
  }

  async updateIntegrityHash(params) {
    try {
      await updaterModel.updateIntegrityHash(params);
    } catch(err) {
      return Promise.reject(err);
    }
  }
}

module.exports = new UpdaterService();
