const versionCheck = require("../../functions/VersionCheck");
const UpdaterService = require('../../service/UpdaterService');
const SystemLogger = require("../../logs/SystemLogger");
const AssistantControl = require("../../service/AssistantControl");
const UpdaterWarning = require('../../helpers/UpdaterWarning');
const Monitor = require('./Monitor');
const SocketRooms = require('../SocketRooms');
const SocketManager = require("../SocketManager");
const PortalEmitter = require("./PortalEmitter");
class Updater {
    async register(socket, { cdfilial, cdcaixa, version, hostname, ip }) {
        socket.CDFILIAL = cdfilial;
        socket.CDCAIXA = cdcaixa || null;
        this.validateVersion(socket, version, cdfilial, cdcaixa);
        try {
            await this.validateOtherConnections(socket)
            this.joinDefaultRoom(socket);
            this.connectedUpdater({ unit: cdfilial, casher: cdcaixa, connected: true, version });
        } catch(error) {
            const duplicateInfo = {};
            duplicateInfo[hostname] = ip;
            UpdaterWarning.addWarning({
                unit: cdfilial,
                casher: cdcaixa,
            }, 'DUPLICATE', duplicateInfo);
            socket.disconnect();
            SystemLogger.write(error);
        }
        socket.on("updaterLogResponse", async (logs) => {
            if(!logs.CDFILIAL) return;
            const {
                masterSocketId,
                usingLatestLines = false,
                ...responseLogs
            } = logs;
            const roomLogsUnit = SocketManager.getRoom(SocketRooms.updater.log(socket.CDFILIAL, socket.CDCAIXA));
            if(usingLatestLines) {
                const targetSocket = await this.getSystemManagementPageSockets({ id: masterSocketId, casher: socket.CDCAIXA });
                // const targetSocket = masterSockets.find( socket => socket.id === masterSocketId);
                if(targetSocket) targetSocket.emit('updaterLog', responseLogs);
            } else {
                roomLogsUnit.emit('updaterLog', responseLogs);
            }
        });

        socket.on("selfUpdateResponseSuccess", async (result) => {
            if (socket.CDCAIXA) {
                result.CDCAIXA = socket.CDCAIXA;
            }
            PortalEmitter.emit('selfUpdateSuccess', result);
        });

        socket.on("selfUpdateResponseError", async (result) => {
            if (socket.CDCAIXA) {
                result.CDCAIXA = socket.CDCAIXA;
            }
            PortalEmitter.emit('selfUpdateError', result);
        });

        socket.on("updaterResultInProgress", async (result) => {
            try {
                const room = this.getUpdateRoomName(socket.CDCAIXA);
                const paramsRoom = SocketManager.getRoom(SocketRooms.master.parametrizationsPage);
                if(result.action !== 'verify') {
                    SocketManager.getRoom(room).emit("updaterInProgress", result);
                }
                switch (result.action) {
                    case 'install':
                        result.payload.OPERATION = 'INSTALACAO_EM_CURSO';
                        break;
                    case 'uninstall':
                        result.payload.OPERATION = 'REMOCAO_EM_CURSO';
                        break;
                    case 'verify':
                        result.payload.OPERATION = 'VERIFICACAO_EM_CURSO';
                        paramsRoom.emit('updaterInProgress', result);
                        break;
                    case 'integrity_analyze':
                        result.payload.OPERATION = 'ANALISE_INTEGRIDADE_EM_CURSO';
                        break;
                }

                await UpdaterService.updateEventTable(result.payload);
            } catch (err) {
                SystemLogger.write(err);
            }
        });

        socket.on("updaterResultFinish", async (result) => {
            try {
                const paramsRoomName = this.getParametrizationRoomName(socket.CDCAIXA);
                const paramsRoom = SocketManager.getRoom(paramsRoomName);
                const roomName = this.getUpdateRoomName(socket.CDCAIXA);
                const room = SocketManager.getRoom(roomName);
                const { CDFILIAL } = socket;
                const socketResponse = { ...result, CDFILIAL };
                const { operation } = socketResponse;
                switch(operation) {
                    case 'VERIFY':
                        if(socketResponse.error) {
                            return paramsRoom.emit('updaterFinish', socketResponse);
                        }
                        const result = await UpdaterService.handleInstalledProducts(socketResponse.payload).catch(() => ({ error: true }));
                        if (result.error) {
                            socketResponse.error = true;
                        } else if (result.invalidProducts) {
                            socketResponse.INVALID_PRODUCTS = result.invalidProducts;
                        }
                        paramsRoom.emit('updaterFinish', socketResponse);
                        break;
                    case 'INTEGRITY_ANALYZE':
                        const analyzeResult = await UpdaterService.updateAnalyzeIntegrity(socketResponse.payload).catch(err => {
                            SystemLogger.write(err);
                            return { error: true };
                        });
                        if (analyzeResult.error) {
                            socketResponse.error = true;
                        }
                        room.emit("updaterFinish", socketResponse);
                        break;
                    case 'INSTALL':
                    case 'UPDATE':
                    case 'REMOVE':
                        room.emit("updaterFinish", socketResponse);
                        const DSHASHANALISE = socketResponse.payload.hash || null;
                        const PRODFILIID = socketResponse.payload.PRODFILIID;
                        if(operation === 'REMOVE') {
                            await UpdaterService.updateResetIntegrityAnalyze(PRODFILIID);
                        } else {
                            await UpdaterService.updateIntegrityHash({ DSHASHANALISE, PRODFILIID });
                        }
                        break;
                }
            } catch(err) {
                SystemLogger.write(err);
            }
        });

        socket.on("updaterChangeServices", (result) => {
            const { CDFILIAL: unit, CDCAIXA: casher = null } = socket;
            switch(result.operation.toUpperCase()) {
                case 'INSTALL':
                    Monitor.addServiceToMonitoring(result.services, { unit, casher });
                    UpdaterService.handleChangeService(unit, casher, result.services, 'INSTALL');
                    break;
                    case 'REMOVE':
                        Monitor.removeServiceToMonitoring(result.services, { unit, casher });
                        UpdaterService.handleChangeService(unit, casher, result.services, 'REMOVE');
                    break;
                default:
                    SystemLogger.write(`OPERATION_NOT_MAPPED ${result.operation}`);
            }
        })

        socket.on('testConnectionResult', (resultTestConn) => {
            const page = [this.getParametrizationRoomName(socket.CDCAIXA)];
            SocketManager.getRoom(page).emit(
                'testConnectionFinish',
                resultTestConn
            );
        });

        socket.on('downloadLogError', ({ error, message, param }) => {
            const { CDFILIAL, CDCAIXA, type } = param;
            const downloadLogRoomName = this.getDownloadLogRoom(CDFILIAL, CDCAIXA, type);
            const room = SocketManager.getRoom(downloadLogRoomName);
            room && room.emit('downloadLogResult', { error, message, param });
        });

        socket.on('serviceUpdaterManagementResult', ({ error, message, param, messageDialog }) => {
            const { CDFILIAL: unit, CDCAIXA: casher = null } = param;
            const { serviceManagement, serviceManagementPerCasher } = SocketRooms.master;
            const room = casher ? serviceManagementPerCasher : serviceManagement;
            const updaterServiceManagementRoom = SocketManager.getRoom(room);
            if (updaterServiceManagementRoom) {
                updaterServiceManagementRoom.emit('serviceManagementResult', { error, message, param, messageDialog });
            }
        });

        socket.on("serviceStatusDetails", ({ services, error = false }) => {
            try {
                const { CDFILIAL, CDCAIXA = null } = socket;
                const { statusServicesManagementAll, statusServicesManagement } = SocketRooms.master;

                const roomAllCasher = statusServicesManagementAll({ CDFILIAL });
                const roomUnitAndCasher  = statusServicesManagement({ CDFILIAL, CDCAIXA });

                const room = SocketManager.getRoom(roomAllCasher).to(roomUnitAndCasher);
                if (room) {
                    room.emit("serviceStatusDetailsResult", { params: { services, unit: CDFILIAL, casher: CDCAIXA }, error });
                }
            } catch (error) {
                SystemLogger.write(error);
            }
        });
    }

    async unregister(socket, { cdfilial, cdcaixa, version }) {
        const socketUpdater = await this.getUpdaterSocketByUnitAndCasher(cdfilial, cdcaixa);

        this.connectedUpdater({
            unit: cdfilial,
            casher: cdcaixa,
            connected: !!socketUpdater,
            version,
        });
    }

    async emit(event, data, { unit, casher = null }) {
        const updaterSocket = await this.getUpdaterSocketByUnitAndCasher(unit, casher);
        if (updaterSocket) {
            updaterSocket.emit(event, data);
        } else {
            throw new Error(`No active socket connection. ${unit}${casher ? ' | ' + casher : ''}`);
        }
    }

    async emitAll(event, data, { unit, casher = null }) {
        const updaterSocket = await this.getUpdaterSocketByUnitAndCasher(unit, casher);

        if (updaterSocket) {
            for (const key in data) {
                updaterSocket.emit(event, data[key]);
            }
        } else {
            throw new Error(
                `No active socket connection. ${unit}${casher ? " | " + casher : ""
                }`
            );
        }
    }

    async executeToMultiSocket({ matchFn, callbackFn }) {
        if(typeof matchFn !== 'function' || typeof callbackFn !== 'function') {
            throw new Error('Invalid type of parameters {matchFn} or {callbackFn}');
        }
        const sockets = await this.getUpdaters();
        sockets.forEach((socket) => {
            if(matchFn(socket)) {
                callbackFn(socket);
            }
        });
    }
    // ---------------------------- //
    getUpdateRoomName(CDCAIXA = null) {
        const {
            updatesPage,
            updatesPerCasherPage
        } = SocketRooms.master;
        return CDCAIXA ? updatesPerCasherPage : updatesPage;
    }
    getParametrizationRoomName(CDCAIXA = null) {
        const {
            parametrizationsPage,
            parametrizationsPerCasherPage
        } = SocketRooms.master;
        return CDCAIXA ? parametrizationsPerCasherPage : parametrizationsPage;
    }

    async connectedUpdater({ unit, casher = null, version = '', connected }) {
        if(!connected) {
            UpdaterWarning.clearWarning({ unit, casher });
        }
        const updateAvaibleVersion = AssistantControl.getUpdateAvailableVersion();
        const { DATA, ...warningStatus } = UpdaterWarning.getWarning({ unit, casher }) || {};
        const activeWarnings = Object.keys(warningStatus).filter(name => warningStatus[name]);
        const socketData = {
            CONNECTED: connected,
            CDFILIAL: unit,
            WARNING: !!activeWarnings.length,
            VERSION: version,
            WARNING_INFO: activeWarnings,
            UPDATE_AVAILABLE: updateAvaibleVersion && connected && updateAvaibleVersion !== version,
            UPDATE_AVAILABLE_VERSION: updateAvaibleVersion || '',
        }
        if(casher) {
            socketData.CDCAIXA = casher;
        }

        PortalEmitter.emit('updaterConnected', socketData);
    }
    joinDefaultRoom( socket ) {
        socket.join(SocketRooms.updater.default);
    }
    joinRooms( socket, isPerCasher = false ) {
        const pages = [SocketRooms.master.default, this.getUpdateRoomName(isPerCasher)];
        socket.join(pages);
    }
    async getUpdaterConnected(CDFILIAL, CDCAIXA = null) {
        const socketUpdater = await this.getUpdaterSocketByUnitAndCasher(CDFILIAL, CDCAIXA);
        const hasSocket = !!socketUpdater;
        return {
            socket: hasSocket ? socketUpdater : null,
            socketInfo: hasSocket ? SocketManager.getSocketHeaderInfo(socketUpdater) : null,
        };
    }
    async validateOtherConnections(socket) {
        const { CDFILIAL, CDCAIXA } = socket;
        const socketUpdater = await this.getUpdaterSocketByUnitAndCasher(CDFILIAL, CDCAIXA);
        const unitConnected = !!socketUpdater && socketUpdater.id !== socket.id;
        if(unitConnected) {
            const { headers = {}, address} = socketUpdater.handshake;
            const { hostname } = headers;
            UpdaterWarning.addWarning({ unit: CDFILIAL, casher: CDCAIXA }, 'DUPLICATE', {
                ACTIVE: { hostname, address }
            });
            const socketData = {
                CDFILIAL,
                WARNING: true,
                WARNING_INFO: ['DUPLICATE'],
            }
            if (CDCAIXA) {
                socketData.CDCAIXA = CDCAIXA;
            }
            PortalEmitter.emit('updaterWarning', socketData);
            const casher = CDCAIXA ? `_${CDCAIXA}` : '';
            let errorMessage = `UNIT_${CDFILIAL}${casher}_IS_ALREADY_CONNECTED`;
            return Promise.reject(new Error(errorMessage));
        }
        return Promise.resolve();
    }
    validateVersion(socket, version, unit, casher) {
        const { MIN_UPDATER_VERSION_SUPPORTED = 'any' } = process.env;
        if (MIN_UPDATER_VERSION_SUPPORTED !== 'any' && versionCheck.greaterThan(MIN_UPDATER_VERSION_SUPPORTED, version)) {
            socket.emit("invalidVersion", {
                supportedUpdaterVersion: MIN_UPDATER_VERSION_SUPPORTED,
                updaterVersion: version,
            });
            UpdaterWarning.addWarning({ unit, casher }, 'INVALID_VERSION', {
                CURRENT_VERSION: version,
                MIN_VERSION: MIN_UPDATER_VERSION_SUPPORTED,
            })
            return false;
        }
        UpdaterWarning.removeWarning({ unit, casher }, 'INVALID_VERSION');
        socket.emit("validVersion");
        return true;
    }
    getDownloadLogRoom(unit, casher, type) {
        const {
            downloadLogTxt,
            downloadLogZip
        } = SocketRooms.master;
        const roomFunction = type === 'txt' ? downloadLogTxt : downloadLogZip;
        const downloadLogRoom = roomFunction.apply(null, [unit, casher]);
        return downloadLogRoom;
    }
    async getUpdaters() {
        return await SocketManager.getRoom(SocketRooms.updater.default).fetchSockets();
    }
    async getSystemManagementPageSockets({ id, casher }){
        const {
            updatesPage,
            updatesPerCasherPage,
        } = SocketRooms.master;
        const room = !!casher ? updatesPerCasherPage : updatesPage;
        const sockets = await SocketManager.getRoom(room).fetchSockets();
        return id ? sockets.find((sckt) => sckt.id === id) : sockets;
    }
    async getUpdaterSocketByUnitAndCasher(CDFILIAL, CDCAIXA = null) {
        const socketUpdaters = await this.getUpdaters();
        const matchSocket = (socket) => {
            return socket.CDFILIAL == CDFILIAL && (!CDCAIXA || socket.CDCAIXA === CDCAIXA);
        };
        const updaterUnitSocket = socketUpdaters.find(matchSocket);
        return updaterUnitSocket;
    }
}

module.exports = new Updater();