const SocketRooms = require('../../SocketRooms');
const Updater = require('../Updater');
const UpdaterLogger = require('../../../logs/UpdaterLogger');
const UpdaterWarning = require('../../../helpers/UpdaterWarning');
const AssistantControl = require('../../../service/AssistantControl');
const PortalEmitter = require('../PortalEmitter');
const SystemLogger = require('../../../logs/SystemLogger');
const UpdaterTelegramData = require('../../../service/UpdaterTelegramData');
const Telegram = require('../../../functions/Telegram');

class SystemManagement {
  register(socket, { page }) {
    const { master } = SocketRooms;
    const roomName =
      page === 'updates'
        ? master.updatesPage
        : master.updatesPerCasherPage;
    socket.join([master.default, roomName]);

    socket.on('getUpdaterLog', async (params) => {
      const data = { ...params, masterSocketId: socket.id };
      const { CDFILIAL: unit, CDCAIXA: casher } = params;
      socket.join(SocketRooms.updater.log(unit, casher));
      Updater.emit('getUpdaterLog', data, { unit, casher }).catch(e => SystemLogger.error(e));
    });

    socket.on('leaveDownloadLogRoom', async (params) => {
      const { CDFILIAL, CDCAIXA, type } = params;
      const roomName = Updater.getDownloadLogRoom(CDFILIAL, CDCAIXA, type);
      socket.leave(roomName);
    });

    socket.on('execUpdater', async (params) => {
      const { CDFILIAL: unit, CDCAIXA: casher } = params;
      Updater.emit('updater', params, { unit, casher }).catch(e => SystemLogger.error(e));
    });

    socket.on('execUpdaterMultiInstall', async (params) => {
      const unitsInfo = params.UNITS.reduce((acc, unitInfo) => {
        const { CDFILIAL, CDCAIXA = null, ...info } = unitInfo;
        const pivot = CDCAIXA ? `${CDFILIAL}_${CDCAIXA}` : CDFILIAL;
        acc[pivot] = info;
        return acc;
      }, {});
      const getSocketPivot = (updaterSocket) => {
        const { CDFILIAL, CDCAIXA } = updaterSocket;
        const pivot = CDCAIXA ? `${CDFILIAL}_${CDCAIXA}` : CDFILIAL;
        return pivot;
      }
      const matchFn = (updaterSocket) => {
        const pivot = getSocketPivot(updaterSocket);
        return (pivot in unitsInfo);
      };
      const callbackFn = (updaterSocket) => {
        const pivot = getSocketPivot(updaterSocket);
        const { PRODFILIID, CURRENTVERSION } = unitsInfo[pivot];
        const DEVOPSID = params.DEVOPSID.toString();
        const socketParams = {
          ...params,
          PRODFILIID,
          CDFILIAL,
          CDPRODUCTID: DEVOPSID,
          DEVOPSID,
          CURRENTVERSION: CURRENTVERSION === '-' ? null : CURRENTVERSION,
        };
        if (updaterSocket.CDCAIXA) {
          socketParams.CDCAIXA = updaterSocket.CDCAIXA;
        }
        updaterSocket.emit('updater', socketParams);
      };
      Updater.executeToMultiSocket({ matchFn, callbackFn });
    });

    socket.on('detailStatusConnection', async (data) => {
      const { CDFILIAL: unit, CDCAIXA: casher = null } = data;
      const STATE = UpdaterWarning.getState({ unit, casher }) || {};
      const { DUPLICATE = {} } = STATE.DATA || {};
      socket.emit('detailStatusConnectionResult', DUPLICATE);
    });

    socket.on('selfUpdateRequest', async (param) => {
      const {
        LATEST,
        NMOPERADOR,
        CDFILIAL: unit,
        NRORG,
        CDCAIXA: casher = null
      } = param;

      try {
        const basicAuth = AssistantControl.getBasicAuth();
        const controlUrl = AssistantControl.getControlUrl();
        const isDevEnv = process.env.NODE_ENV === 'development';
        const data = {
          ip: isDevEnv ? 'Ambiente de Desenvolvimento' : socket.handshake.address,
          user: NMOPERADOR,
          CDFILIAL: unit,
          NRORG,
        };
        const telegramData = await UpdaterTelegramData.getData(data, 'self-updater');
        Telegram.sendNotification('telegram-update-system', telegramData);
        // Envia requisição de atualização para a loja
        const socketData = {
          OPERATOR: NMOPERADOR,
          NEWVERSION: LATEST,
          URL: `${controlUrl}/updater-version/download`,
          AUTH_TOKEN: `Basic ${basicAuth}`,
        };
        await Updater.emit('selfUpdate', socketData, { unit, casher });
      } catch(error) {
        SystemLogger.error(error);
      }
    });

    socket.on('downloadLog', async (param) => {
      try {
        const {
          CDFILIAL: unit,
          CDCAIXA: casher = null,
          PRODUCT,
          type
        } = param;

        const { fileName, useLocalLog } = await UpdaterLogger.getLocalLogInfo(
          unit,
          casher,
          PRODUCT
        );

        if (useLocalLog && type === 'zip') {
          param.downloadUrl = `/updater-logs/${fileName}`;
          socket.emit('downloadLogResult', { error: false, param });
          return;
        }

        const downloadLogRoom = Updater.getDownloadLogRoom(unit, casher, type);

        socket.join(downloadLogRoom);
        await Updater.emit('downloadLog', param, { unit, casher });
      } catch (error) {
        SystemLogger.error(error);
      }
    });

    socket.on('removeWarning', async (data) => {
      const { CDFILIAL: unit, CDCAIXA: casher, WARNING_INFO } = data;
      UpdaterWarning.removeWarning({ unit, casher }, WARNING_INFO);
      const socketUpdater = Updater.getUpdaterSocketByUnitAndCasher(unit, casher);
      const latestVersion = AssistantControl.getUpdateAvailableVersion();
      const warnings = UpdaterWarning.getActiveWarnings({ unit, casher });
      const connected = !!socketUpdater;
      const socketData = {
        CDFILIAL: unit,
        CDCAIXA: casher,
        CONNECTED: connected,
        WARNING: !!warnings.length,
        WARNING_INFO: warnings,
        UPDATE_AVAILABLE:
          connected && latestVersion && socketUpdater.version !== latestVersion,
        UPDATE_AVAILABLE_VERSION: latestVersion,
      }
      PortalEmitter.emit('updaterConnected', socketData);
    });
  }
}

module.exports = new SystemManagement();
